package meerkat.mixer.proofs;

import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.protobuf.ConcreteCrypto;
import meerkat.protobuf.Crypto;
import org.bouncycastle.math.ec.ECPoint;
import org.factcenter.qilin.primitives.concrete.ECGroup;

/**
 * use for organize the input for each ZKP
 *
 * both meerkat.mixer.proofs and meerkat.mixer.verifier implementation use it
 */
public class ECElGamalMixParams {

    private final ECElGamalEncryption encryptor;
    private final ECGroup group;
    private final ECPoint g;
    private final ECPoint h;

    // Cache encoded formats to save multiple re-encodings
    private final ConcreteCrypto.GroupElement gEncoded;
    private final ConcreteCrypto.GroupElement hEncoded;

    /**
     * @param encryptor encryptor used for encoding/decoding serialized ciphertexts. The group, default generator and
     *                  second base (h) are taken from the encryptor (second base is the public key)
     */
    public ECElGamalMixParams(ECElGamalEncryption encryptor){
        this.encryptor = encryptor;
        this.group = encryptor.getGroup();
        this.g = group.getGenerator();
        this.h = encryptor.getElGamalPK().getPK();
        this.gEncoded = encryptor.encodeElement(g);
        this.hEncoded = encryptor.encodeElement(h);
    }

    public enum TrueCouple {
        left, right, unknown
    }


    /**
     * can be used by anyone, e.g meerkat.mixer.verifier
     *
     * call to the meerkat.mixer.main overload with flag = false
     */
    public Statement createStatement(Crypto.RerandomizableEncryptedMessage in1, Crypto.RerandomizableEncryptedMessage in2
            , Crypto.RerandomizableEncryptedMessage out1, Crypto.RerandomizableEncryptedMessage out2) throws InvalidProtocolBufferException {

        ConcreteCrypto.ElGamalCiphertext in1ElGamal = ECElGamalEncryption.RerandomizableEncryptedMessage2ElGamalCiphertext(in1);
        ConcreteCrypto.ElGamalCiphertext in2ElGamal = ECElGamalEncryption.RerandomizableEncryptedMessage2ElGamalCiphertext(in2);
        ConcreteCrypto.ElGamalCiphertext out1ElGamal = ECElGamalEncryption.RerandomizableEncryptedMessage2ElGamalCiphertext(out1);
        ConcreteCrypto.ElGamalCiphertext out2ElGamal = ECElGamalEncryption.RerandomizableEncryptedMessage2ElGamalCiphertext(out2);
        return new Statement(in1ElGamal, in2ElGamal, out1ElGamal, out2ElGamal);
    }

    /**

     * convert each encrypted message to ElGamalCiphertext
     *
     * @throws InvalidProtocolBufferException - in case that at least one of the encrypted messages isn't
     *                                          ElGamalCiphertext
     */
    protected ProverStatement createProverStatement(Crypto.RerandomizableEncryptedMessage in1, Crypto.RerandomizableEncryptedMessage in2
            , Crypto.RerandomizableEncryptedMessage out1, Crypto.RerandomizableEncryptedMessage out2
            , Crypto.EncryptionRandomness r1, Crypto.EncryptionRandomness r2, boolean switched)
            throws InvalidProtocolBufferException {

        //convert RerandomizableEncryptedMessage to ElGamalCiphertext
        ConcreteCrypto.ElGamalCiphertext in1ElGamal = ECElGamalEncryption.RerandomizableEncryptedMessage2ElGamalCiphertext(in1);
        ConcreteCrypto.ElGamalCiphertext in2ElGamal = ECElGamalEncryption.RerandomizableEncryptedMessage2ElGamalCiphertext(in2);
        ConcreteCrypto.ElGamalCiphertext out1ElGamal = ECElGamalEncryption.RerandomizableEncryptedMessage2ElGamalCiphertext(out1);
        ConcreteCrypto.ElGamalCiphertext out2ElGamal = ECElGamalEncryption.RerandomizableEncryptedMessage2ElGamalCiphertext(out2);

        return new ProverStatement(in1ElGamal, in2ElGamal, out1ElGamal, out2ElGamal, r1, r2, switched);
    }


    /**
     * Statement to be proved.
     *
     * The actual stored data is a cached representation for use in proofs and verification. This consists of the four substatements that are ORs of the DLOG equality.
     *
     * A Statement can be constructed only by calling the {@link #createStatement} factory method on an instance of ECElGamalMixParams  (all constructors are private)
     *
     */
    public class Statement {

        /**
         * Denote the first ciphertext pair e1=(e1c1,e1c2), e2=(e2c1,e2c2) and the second
         * e1N=(e1Nc1, e1Nc2) and e2N=(e2Nc1,e2Nc2).
         * Then ci_ejNDivek = ejNci/ekci
         */
        ECPoint c1_e1NDive1;
        /**
         * See {@link #c1_e1NDive1}
         */
        ECPoint c1_e2NDive1;

        /**
         * See {@link #c1_e1NDive1}
         */
        ECPoint c1_e1NDive2;

        /**
         * See {@link #c1_e1NDive1}
         */
        ECPoint c1_e2NDive2;

        /**
         * See {@link #c1_e1NDive1}
         */
        ECPoint c2_e1NDive1;

        /**
         * See {@link #c1_e1NDive1}
         */
        ECPoint c2_e2NDive1;

        /**
         * See {@link #c1_e1NDive1}
         */
        ECPoint c2_e1NDive2;

        /**
         * See {@link #c1_e1NDive1}
         */
        ECPoint c2_e2NDive2;


        private final OrStatement orStatements[] = new OrStatement[4];


        public OrStatement getOrStatement(int num) {
            if (orStatements[0] == null)
                generateOrStatements();

            return orStatements[num];
        }

        public void setOrStatement(int num, OrStatement orStatement) {
            orStatements[num] = orStatement;
        }



        /**
         * Generate and set the four substatements.
         */
        protected void generateOrStatements() {
            setOrStatement(0, new OrStatement(c1_e1NDive1,c2_e1NDive1,c1_e1NDive2,c2_e1NDive2));
            setOrStatement(1, new OrStatement(c1_e1NDive1,c2_e1NDive1,c1_e2NDive1,c2_e2NDive1));
            setOrStatement(2, new OrStatement(c1_e1NDive2,c2_e1NDive2,c1_e2NDive2,c2_e2NDive2));
            setOrStatement(3, new OrStatement(c1_e2NDive1,c2_e2NDive1,c1_e2NDive2,c2_e2NDive2));
        }


        private Statement(ConcreteCrypto.ElGamalCiphertext e1, ConcreteCrypto.ElGamalCiphertext e2,
                          ConcreteCrypto.ElGamalCiphertext e1New, ConcreteCrypto.ElGamalCiphertext e2New){

            ECPoint e1c1 = encryptor.decodeElement(e1.getC1());
            ECPoint e1c2 = encryptor.decodeElement(e1.getC2());
            ECPoint e2c1 = encryptor.decodeElement(e2.getC1());
            ECPoint e2c2 = encryptor.decodeElement(e2.getC2());
            ECPoint e1Nc1 = encryptor.decodeElement(e1New.getC1());
            ECPoint e1Nc2 = encryptor.decodeElement(e1New.getC2());
            ECPoint e2Nc1 = encryptor.decodeElement(e2New.getC1());
            ECPoint e2Nc2 = encryptor.decodeElement(e2New.getC2());

            c1_e1NDive1 = group.add(e1Nc1, group.negate(e1c1));
            c1_e2NDive1 = group.add(e2Nc1, group.negate(e1c1));
            c1_e1NDive2 = group.add(e1Nc1, group.negate(e2c1));
            c1_e2NDive2 = group.add(e2Nc1, group.negate(e2c1));
            c2_e1NDive1 = group.add(e1Nc2, group.negate(e1c2));
            c2_e2NDive1 = group.add(e2Nc2, group.negate(e1c2));
            c2_e1NDive2 = group.add(e1Nc2, group.negate(e2c2));
            c2_e2NDive2 = group.add(e2Nc2, group.negate(e2c2));
        }
    }

    /**
     * A statement with additional witness information that enables us to generate a ZK proof.
     */
    public class ProverStatement extends Statement {
        /**
         * True iff the ciphertexts were switched (i.e., decrypt(e1N) == decrypt(e2) and decrypt(e2N) == decrypt(e1)
         */
        boolean switched;

        /**
         * Encryption randomness for e1 rerandomization
         */
        Crypto.EncryptionRandomness r1;

        /**
         * Encryption randomnesss for e2 rerandomization
         */
        Crypto.EncryptionRandomness r2;

        @Override
        protected void generateOrStatements() {
            ConcreteCrypto.GroupElement c1_e1NDive1Encoded = encryptor.encodeElement(c1_e1NDive1);
            ConcreteCrypto.GroupElement c1_e2NDive1Encoded = encryptor.encodeElement(c1_e2NDive1);
            ConcreteCrypto.GroupElement c1_e1NDive2Encoded = encryptor.encodeElement(c1_e1NDive2);
            ConcreteCrypto.GroupElement c1_e2NDive2Encoded = encryptor.encodeElement(c1_e2NDive2);
            ConcreteCrypto.GroupElement c2_e1NDive1Encoded = encryptor.encodeElement(c2_e1NDive1);
            ConcreteCrypto.GroupElement c2_e2NDive1Encoded = encryptor.encodeElement(c2_e2NDive1);
            ConcreteCrypto.GroupElement c2_e1NDive2Encoded = encryptor.encodeElement(c2_e1NDive2);
            ConcreteCrypto.GroupElement c2_e2NDive2Encoded = encryptor.encodeElement(c2_e2NDive2);


            if (!switched) {
                setOrStatement(0, new OrProverStatement(c1_e1NDive1, c2_e1NDive1, c1_e1NDive2, c2_e1NDive2
                        , c1_e1NDive1Encoded, c2_e1NDive1Encoded, c1_e1NDive2Encoded, c2_e1NDive2Encoded
                        , r1, TrueCouple.left));
                setOrStatement(1, new OrProverStatement(c1_e1NDive1, c2_e1NDive1, c1_e2NDive1, c2_e2NDive1
                        , c1_e1NDive1Encoded, c2_e1NDive1Encoded, c1_e2NDive1Encoded, c2_e2NDive1Encoded
                        , r1, TrueCouple.left));
                setOrStatement(2, new OrProverStatement(c1_e1NDive2, c2_e1NDive2, c1_e2NDive2, c2_e2NDive2
                        , c1_e1NDive2Encoded, c2_e1NDive2Encoded, c1_e2NDive2Encoded, c2_e2NDive2Encoded
                        , r2, TrueCouple.right));
                setOrStatement(3, new OrProverStatement(c1_e2NDive1, c2_e2NDive1, c1_e2NDive2, c2_e2NDive2
                        , c1_e2NDive1Encoded, c2_e2NDive1Encoded, c1_e2NDive2Encoded, c2_e2NDive2Encoded
                        , r2, TrueCouple.right));
            } else {
                setOrStatement(0, new OrProverStatement(c1_e1NDive1, c2_e1NDive1, c1_e1NDive2, c2_e1NDive2
                        , c1_e1NDive1Encoded, c2_e1NDive1Encoded, c1_e1NDive2Encoded, c2_e1NDive2Encoded
                        , r2, TrueCouple.right));
                setOrStatement(1, new OrProverStatement(c1_e1NDive1, c2_e1NDive1, c1_e2NDive1, c2_e2NDive1
                        , c1_e1NDive1Encoded, c2_e1NDive1Encoded, c1_e2NDive1Encoded, c2_e2NDive1Encoded
                        , r1, TrueCouple.right));
                setOrStatement(2, new OrProverStatement(c1_e1NDive2, c2_e1NDive2, c1_e2NDive2, c2_e2NDive2
                        , c1_e1NDive2Encoded, c2_e1NDive2Encoded, c1_e2NDive2Encoded, c2_e2NDive2Encoded
                        , r2, TrueCouple.left));
                setOrStatement(3, new OrProverStatement(c1_e2NDive1, c2_e2NDive1, c1_e2NDive2, c2_e2NDive2
                        , c1_e2NDive1Encoded, c2_e2NDive1Encoded, c1_e2NDive2Encoded, c2_e2NDive2Encoded
                        , r1, TrueCouple.left));
            }
        }


        @Override
        public OrProverStatement getOrStatement(int num) {
            OrStatement orStatement = super.getOrStatement(num);
            if (!(orStatement instanceof  OrProverStatement))
                throw new RuntimeException("ProverStatement should always have OrProverStatement substatements!!");

            return (OrProverStatement) orStatement;
        }

        private ProverStatement(ConcreteCrypto.ElGamalCiphertext e1, ConcreteCrypto.ElGamalCiphertext e2,
                                ConcreteCrypto.ElGamalCiphertext e1New, ConcreteCrypto.ElGamalCiphertext e2New,
                                Crypto.EncryptionRandomness r1, Crypto.EncryptionRandomness r2, boolean switched) {
            super(e1, e2, e1New, e2New);
            this.r1 = r1;
            this.r2 = r2;
            this.switched = switched;
        }

    }

    /**
     * The statement of a single disjunction to be proved:
     * Either there exists x s.t (g1 ^ x == h1 and g2 ^ x == h2) or there exists x s.t. (g1' ^ x == h1 and g2' ^ x == h2)
     *
     * The statement can't be constructed externally (all constructors are private)
     *
     * 4 instances will be constructed for each new {@link ECElGamalMixParams.Statement}
     *
     */
    public class OrStatement {
        public final ECPoint g1;
        public final ECPoint h1;
        public final ECPoint g2;
        public final ECPoint h2;
        public final ECPoint g1Tag;
        public final ECPoint h1Tag;
        public final ECPoint g2Tag;
        public final ECPoint h2Tag;


        /**
         * used by meerkat.mixer.proofs only
         */
        private OrStatement(ECPoint h1, ECPoint h2, ECPoint h1Tag, ECPoint h2Tag) {
            this.g1 = g;
            this.h1 = h1;
            this.g2 = h;
            this.h2 = h2;
            this.g1Tag = g;
            this.h1Tag = h1Tag;
            this.g2Tag = h;
            this.h2Tag = h2Tag;



        }
    }

    /**
     * An {@link OrStatement} with additional info needed to generate a ZK proof.
     */
    public class OrProverStatement extends OrStatement {

        protected final Crypto.EncryptionRandomness x;
        protected final TrueCouple flag;

        // We cache the encoded versions since we need them as input to the random oracle
        // when using the Fiat-Shamir heuristic.

        protected final ConcreteCrypto.GroupElement g1Encoded;
        protected final ConcreteCrypto.GroupElement h1Encoded;
        protected final ConcreteCrypto.GroupElement g2Encoded;
        protected final ConcreteCrypto.GroupElement h2Encoded;

        protected final ConcreteCrypto.GroupElement g1TagEncoded;
        protected final ConcreteCrypto.GroupElement h1TagEncoded;
        protected final ConcreteCrypto.GroupElement g2TagEncoded;
        protected final ConcreteCrypto.GroupElement h2TagEncoded;

        private OrProverStatement(ECPoint h1, ECPoint h2, ECPoint h1Tag, ECPoint h2Tag,
                                  ConcreteCrypto.GroupElement h1Encoded, ConcreteCrypto.GroupElement h2Encoded, ConcreteCrypto.GroupElement h1TagEncoded, ConcreteCrypto.GroupElement h2TagEncoded,
                                  Crypto.EncryptionRandomness x, TrueCouple flag) {
            super(h1, h2, h1Tag, h2Tag);
            this.g1Encoded = gEncoded;
            this.h1Encoded = h1Encoded;
            this.g2Encoded = hEncoded;
            this.h2Encoded = h2Encoded;
            this.g1TagEncoded = gEncoded;
            this.h1TagEncoded = h1TagEncoded;
            this.g2TagEncoded = hEncoded;
            this.h2TagEncoded = h2TagEncoded;

            this.x = x;
            this.flag = flag;
        }
    }
}
