package meerkat.mixer.proofs;

import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.crypto.concrete.Util;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeProver;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Mixing;
import org.bouncycastle.math.ec.ECPoint;
import org.factcenter.qilin.primitives.RandomOracle;
import org.factcenter.qilin.primitives.concrete.ECGroup;

import java.math.BigInteger;
import java.util.Random;


/**
 * implements of Mix2ZeroKnowledgeProver interface
 * this implementation assumes that each RerandomizableEncryptedMessage is ElGamalCiphertext
 */
public class Prover implements Mix2ZeroKnowledgeProver {

    private final ECGroup group;
    private final RandomOracle randomOracle;
    private final Random rand;
    private final ECElGamalEncryption encryptor;
    private final ECPoint g,h;
    private final BigInteger groupOrderUpperBound;
    private final ECElGamalMixParams mixParams;

    /**
     * @param rand
     * @param encryptor
     * @param randomOracle  - use for Fiat–Shamir heuristic
     */
    public Prover(Random rand, ECElGamalEncryption encryptor, RandomOracle randomOracle) {

        this.rand = rand;
        this.encryptor = encryptor;
        this.randomOracle = randomOracle;
        this.group = this.encryptor.getGroup();
        this.g = group.getGenerator();
        this.h = this.encryptor.getElGamalPK().getPK();
        this.mixParams = new ECElGamalMixParams(encryptor);
        this.groupOrderUpperBound  = group.orderUpperBound();
    }

    /**
     * @param in1
     * @param in2
     * @param out1  - if switched then out1 = rerandomize(in2,r2) else out1 = rerandomize(in1,r1)
     * @param out2  - if switched then out2 = rerandomize(in1,r1) else out1 = rerandomize(in2,r2)
     * @param switched    - flag
     * @param i     - column of in1 and out1 in encryption table
     * @param j     - column of in2 and out2 in encryption table
     * @param layer - row of in1,in2 in encryption table
     * @param r1
     * @param r2
     * @return      - a valid ZKP that indeed out1,out2 calculated as required
     * @throws InvalidProtocolBufferException
     */
    public Mixing.ZeroKnowledgeProof prove(Crypto.RerandomizableEncryptedMessage in1,
                                           Crypto.RerandomizableEncryptedMessage in2,
                                           Crypto.RerandomizableEncryptedMessage out1,
                                           Crypto.RerandomizableEncryptedMessage out2,
                                           boolean switched,int i,int j, int layer,
                                           Crypto.EncryptionRandomness r1,
                                           Crypto.EncryptionRandomness r2) throws InvalidProtocolBufferException {
        Mixing.ZeroKnowledgeProof.OrProof first,second,third,fourth;

        ECElGamalMixParams.ProverStatement statement = mixParams.createProverStatement(in1,in2,out1,out2,r1,r2,switched);

        first = createOrProof(statement.getOrStatement(0));
        second = createOrProof(statement.getOrStatement(1));
        third = createOrProof(statement.getOrStatement(2));
        fourth = createOrProof(statement.getOrStatement(3));

        Mixing.ZeroKnowledgeProof.Location location = Mixing.ZeroKnowledgeProof.Location.newBuilder()
                .setI(i)
                .setJ(j)
                .setLayer(layer)
                .build();

        Mixing.ZeroKnowledgeProof result = Mixing.ZeroKnowledgeProof.newBuilder()
                .setFirst(first)
                .setSecond(second)
                .setThird(third)
                .setFourth(fourth)
                .setLocation(location)
                .build();
        return  result;
    }


    /**
     * Fiat–Shamir heuristic
     * @param input - protobuf contains all parameters from the first step of the current proof
     * @param randomOracle
     * @return randomOracle.hash(input)
     */
    public static BigInteger hash(Mixing.ZeroKnowledgeProof.OrProof.FirstMessage input, RandomOracle randomOracle) {
        byte[] arr = input.toByteArray();
        return new BigInteger(1,randomOracle.hash(arr,arr.length));
    }


    /**
     * Generate a ZK proof that there exists x s.t (g1 ^ x == h1 and g2 ^ x == h2) or (g1' ^ x == h1 and g2' ^ x == h2).
     *
     * For each clause of the disjunction, we use the following sigma-protocol for DLOG equality (i.e. log_{g1}(h1)==log_{g1}(h2)):
     * <ol>
     *     <li>Prover chooses a random r, and sends g1^r, g2^r </li>
     *     <li>Verifier chooses a random c and sends c</li>
     *     <li>Prover computes </li>
     * </ol>
     *
     *
     * @param orStatement
     * @return ZKP OrProof:
     *         assuming DLog is hard in this.group then that proves x is known for the meerkat.mixer.proofs
     */
    private Mixing.ZeroKnowledgeProof.OrProof createOrProof(ECElGamalMixParams.OrProverStatement orStatement) {

        ECPoint g1 = orStatement.g1;
        ECPoint h1 = orStatement.h1;
        ECPoint g2 = orStatement.g2;
        ECPoint h2 = orStatement.h2;

        ECPoint g1Tag  = orStatement.g1Tag;
        ECPoint h1Tag  = orStatement.h1Tag;
        ECPoint g2Tag  = orStatement.g2Tag;
        ECPoint h2Tag  = orStatement.h2Tag;

        // Randomness for the ZK proof
        BigInteger r = encryptor.generateRandomExponent(rand);


        BigInteger c1,c2,z,zTag;
        ECPoint u,v,uTag,vTag;
        Mixing.ZeroKnowledgeProof.OrProof.FirstMessage firstMessage;


        switch (orStatement.flag) {
            case left:
                c2 = encryptor.generateRandomExponent(rand);
                zTag = encryptor.generateRandomExponent(rand);
                //step 1
                u = group.multiply(g1, r);
                v = group.multiply(g2, r);
                uTag = group.add(group.multiply(g1Tag, zTag), group.negate(group.multiply(h1Tag, c2)));
                vTag = group.add(group.multiply(g2Tag, zTag), group.negate(group.multiply(h2Tag, c2)));
                //step 2
                // c1 = (hash(input + step1) + group size - c2)% group size
                firstMessage =
                        Mixing.ZeroKnowledgeProof.OrProof.FirstMessage.newBuilder()
                                .setG1(orStatement.g1Encoded)
                                .setH1(orStatement.h1Encoded)
                                .setG2(orStatement.g2Encoded)
                                .setH2(orStatement.h2Encoded)
                                .setG1Tag(orStatement.g1TagEncoded)
                                .setH1Tag(orStatement.h1TagEncoded)
                                .setG2Tag(orStatement.g2TagEncoded)
                                .setH2Tag(orStatement.h2TagEncoded)
                                .setU(encryptor.encodeElement(u))
                                .setV(encryptor.encodeElement(v))
                                .setUTag(encryptor.encodeElement(uTag))
                                .setVTag(encryptor.encodeElement(vTag))
                                .build();
                c1 = hash(firstMessage,randomOracle).add(group.orderUpperBound().subtract(c2)).mod(groupOrderUpperBound);
                //step 3
                //z = (r +  c1 * x) % group size;
                z = r.add(c1.multiply(encryptor.extractRandomness(orStatement.x))).mod(groupOrderUpperBound);
                break;
            case right:
                c1 = encryptor.generateRandomExponent(rand);
                z = encryptor.generateRandomExponent(rand);
                //step 1
                uTag = group.multiply(g1Tag, r);
                vTag = group.multiply(g2Tag, r);
                u = group.add(group.multiply(g1, z), group.negate(group.multiply(h1, c1)));
                v = group.add(group.multiply(g2, z), group.negate(group.multiply(h2, c1)));
                //step 2
                // c1 = (hash(input + step1) + group size - c1)% group size
                firstMessage =
                        Mixing.ZeroKnowledgeProof.OrProof.FirstMessage.newBuilder()
                                .setG1(orStatement.g1Encoded)
                                .setH1(orStatement.h1Encoded)
                                .setG2(orStatement.g2Encoded)
                                .setH2(orStatement.h2Encoded)
                                .setG1Tag(orStatement.g1TagEncoded)
                                .setH1Tag(orStatement.h1TagEncoded)
                                .setG2Tag(orStatement.g2TagEncoded)
                                .setH2Tag(orStatement.h2TagEncoded)
                                .setU(encryptor.encodeElement(u))
                                .setV(encryptor.encodeElement(v))
                                .setUTag(encryptor.encodeElement(uTag))
                                .setVTag(encryptor.encodeElement(vTag))
                                .build();
                c2 = hash(firstMessage,randomOracle).add(group.orderUpperBound().subtract(c1)).mod(groupOrderUpperBound);
                //step 3
                //zTag = (r +  c2 * x) % group size;
                zTag = r.add(c2.multiply(encryptor.extractRandomness(orStatement.x))).mod(groupOrderUpperBound);
                break;
            default:
                return null;
        }


        return Mixing.ZeroKnowledgeProof.OrProof.newBuilder()
                .setG1(firstMessage.getG1())
                .setH1(firstMessage.getH1())
                .setG2(firstMessage.getG2())
                .setH2(firstMessage.getH2())
                .setG1Tag(firstMessage.getG1())
                .setH1Tag(firstMessage.getH1Tag())
                .setG2Tag(firstMessage.getG2Tag())
                .setH2Tag(firstMessage.getH2Tag())
                .setU(firstMessage.getU())
                .setV(firstMessage.getV())
                .setUTag(firstMessage.getUTag())
                .setVTag(firstMessage.getVTag())
                .setC1(Util.encodeBigInteger(c1))
                .setC2(Util.encodeBigInteger(c2))
                .setZ(Util.encodeBigInteger(z))
                .setZTag(Util.encodeBigInteger(zTag))
                .build();
    }
}

